'''
Author: Pengyao PING
Date: 2022-08-14 14:40:10
LastEditors: Pengyao PING
LastEditTime: 2022-10-24 13:12:03
Email: Pengyao.Ping@student.uts.edu.au
Description: 
'''
import collections
from Bio import SeqIO
# import gzip
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from tqdm import tqdm
import copy
import random
from mpire import WorkerPool
import os
from SRRec.utils import *
from SRRec.data_generation import DataGneration
from SRRec.error_orrection import ErrorCorrection
from SRRec.data_preprocessing import DataProcessing

class Simulation():
    def __init__(self, logger, num_workers, f_in, error_rate, output_dir, substations, indels, seed=0):
        self.logger = logger
        self.num_workers = num_workers

        self.f_in = f_in
        self.error_rate = error_rate
        self.substations = substations
        self.indels = indels
        self.seed = seed
        self.output_dir = output_dir
        if os.path.exists(self.output_dir):
            self.logger.info("Directory '% s' already exists" % self.output_dir)
            # for f in os.listdir(self.output_dir):
            #     os.remove(os.path.join(self.output_dir, f))
        else:
            os.makedirs(self.output_dir)
            self.logger.info("Directory '% s' created" % self.output_dir)

    def enumerate_ed1_seqs(self, read):
        possible_ed1 = []
        if self.substations:
            possible_ed1.extend(self.seq2substitution(read))
        if self.indels:
            possible_ed1.extend(self.seq2deletion(read))
            possible_ed1.extend(self.seq2insertion(read))
        return list(set(possible_ed1))

    def replace_char(self, seq, char, index):
        seq[index] = char
        return ''.join(seq)

    def sub_base(self, base):
        if base == 'A':
            return 'TCG'
        elif base == 'T':
            return 'ACG'
        elif base == 'C':
            return 'ATG'
        elif base == 'G':
            return 'ACT'
        elif base == 'N':
            return 'N'
            
    def seq2substitution(self, read):
        """
        enumerate all the substitutions for one read 
        Args:
            read (str): a sequence
        Returns:
            set: a set contains reads
        """
        editdis1_list = []
        # substitution
        raw_seq = list(read)
        n = len(raw_seq)
        for i in range(n):
            seq = copy.deepcopy(raw_seq)
            temp_base = seq[i]
            #print(temp_base)
            temp_sub_base = list(self.sub_base(temp_base))
            #print(temp_sub_base)
            for b in temp_sub_base:
                if b != 'N':
                    sub_seq = self.replace_char(seq, b, i)
                    editdis1_list.append(sub_seq)
        return set(editdis1_list)

    def seq2deletion(self, read):
        """
        enumerate all the deletions for one read 
        Args:
            read (str): a sequence
        Returns:
            set: a set contains reads
        """
        editdis1_list = []
        seq = list(read)
        n = len(seq)
        # deletion
        for i in range(n):
            del_seq = read[:i] + read[(i+1):]
            editdis1_list.append(del_seq)
        return set(editdis1_list)

    def seq2insertion(self, read):
        """
        enumerate all the insertions for one read 
        Args:
            read (str): a sequence
        Returns:
            set: a set contains reads
        """
        editdis1_list = []
        seq = list(read)
        n = len(seq)
        # insertion
        bases = ['A', 'G', 'C', 'T']
        for i in range(n+1):
            for b in bases:
                raw_seq = copy.deepcopy(seq)
                raw_seq.insert(i, b)
                editdis1_list.append(''.join(raw_seq))
        return set(editdis1_list)
        
    def read2err_read(self, read):
        possible_seqs = self.enumerate_ed1_seqs(read)
        num = len(possible_seqs)
        # random.seed(self.seed)
        random.shuffle(possible_seqs)
        idx = random.randint(0, num-1)
        return possible_seqs[idx]

    def error_correction(self):
        DG = DataGneration(
            self.logger,
            self.num_workers,
            False,
            self.f_in, 
            self.output_dir + "DG/",
            high_freq_thre=5,
            max_error_freq=3,
            label0_min_num=10000,
            save_graph=False,
            graph_visualization=False,
            drawing_graph_num=50
            )

        labelled_csv, max_read_len, isolates_file, non_isolates_file, errorTypes = DG.data_files()
        ###############################################################
        EC = ErrorCorrection(
            self.logger,
            errorTypes,
            self.num_workers,
            self.f_in, 
            self.output_dir + "DG/",
            self.output_dir + "RV/", 
            self.output_dir + "output/", 
            ambiguous_prediction=False)
        corrected_file = EC.Correction(isolates_file, non_isolates_file)
        ##############################################################
        update_corrected_file = DG.ed2_correction(corrected_file, self.output_dir + "output/")
        return update_corrected_file        

    def simulation(self):
        # step 1 correct errors using SRRec
        # corrected_data = self.error_correction()
        # inject errors
        # raw_data, true_data = self.error_injection(corrected_data)
        raw_data, true_data = self.error_injection(self.f_in)
        DP = DataProcessing( 
            self.logger,
            self.num_workers,
            self.output_dir)
        umi_raw_dataset, umi_true_dataset = DP.write_mimic_umis(raw_data, true_data)
        return umi_raw_dataset, umi_true_dataset
        
    def error_injection(self, data_set):
        # records, file_type = self.parse_data_index(self.f_in)
        records_dict, file_type = parse_data_dict(data_set)
        id_lst = list(records_dict.keys())

        total_seqs_lst = []
        seq2id_dict = {}
        for id in id_lst:
            seq = str(records_dict[id].seq)
            seq2id_dict.setdefault(seq, []).append(id)
            total_seqs_lst.append(seq)
            # total_bases += len(seq)
            # total_reads += 1

        read2counts = collections.Counter(total_seqs_lst)
        select_id_lst = []
        total_bases = 0
        total_reads = 0
        new_read2counts = {}
        for read, count in read2counts.items():
            if count >= 5:
                new_read2counts[read] = count
                select_id_lst.extend(seq2id_dict[read])
                total_reads += count
                total_bases += len(read) * count

        err_reads_num = round(total_bases * self.error_rate)
    
        if total_reads > err_reads_num:
            per_num =  err_reads_num / total_reads
        err_id_lst = []
        for read, count in new_read2counts.items():
            random_num = round(count * per_num)
            name_lst = seq2id_dict[read]
            random.shuffle(name_lst)
            err_id_lst.extend(random.sample(name_lst, random_num))

        # err_reads_num = round(total_bases * self.error_rate / total_reads)
        self.logger.info(f"total bases: {total_bases}, total reads: {total_reads}, error rate: {self.error_rate}, calculated error reads number: {err_reads_num}, actual error reads number: {len(err_id_lst)}")

        err_subdataset = self.output_dir + "err_subdataset" + file_type
        err_records = []
        shared_objects = records_dict, file_type
        with WorkerPool(self.num_workers, shared_objects=shared_objects, start_method='fork') as pool:
            with tqdm(total=len(err_id_lst), desc=self.logger.info("Generate Mutated Seqs")) as pbar:   
                for tmp_rec in pool.imap(self.mutate_seq, err_id_lst):
                    # print(tmp_rec)
                    err_records.append(tmp_rec)
                    pbar.update() 
        # print(type(sub_records[0]), sub_records[0])
        with open(err_subdataset, "w") as handle:
            SeqIO.write(err_records, handle, file_type)
        ########################################################################################
        err_free_subdataset = self.output_dir + "err_free_subdataset" + file_type
        err_free_id_lst = list(set(select_id_lst) - set(err_id_lst))
        err_free_records = []
        with WorkerPool(self.num_workers, shared_objects=shared_objects, start_method='fork') as pool:
            with tqdm(total=len(err_free_id_lst), desc=self.logger.info("Extract Seqs")) as pbar:   
                for tmp_rec in pool.imap(self.extract_seq, err_free_id_lst):
                    # print(tmp_rec)
                    err_free_records.append(tmp_rec)
                    pbar.update() 
        with open(err_free_subdataset, "w") as handle:
            SeqIO.write(err_free_records, handle, file_type)
        base = self.f_in.split("/")[-1]
        file_name = base.split(file_type)[0]
        raw_data = self.output_dir + "simulated_raw_" + file_name + file_type
        os.system("cat %s %s > %s" % (err_subdataset, err_free_subdataset, raw_data))
        ##################################################################################
        true_data = self.output_dir + "simulated_true_" + file_name + file_type
        err_free_records_2 = []
        with WorkerPool(self.num_workers, shared_objects=shared_objects, start_method='fork') as pool:
            with tqdm(total=len(err_id_lst), desc=self.logger.info("Extract Seqs")) as pbar:   
                for tmp_rec in pool.imap(self.extract_seq, err_id_lst):
                    # print(tmp_rec)
                    err_free_records_2.append(tmp_rec)
                    pbar.update() 
        err_free_subdataset_2 = self.output_dir + "err_free_subdataset2" + file_type
        with open(err_free_subdataset_2, "w") as handle:
            SeqIO.write(err_free_records_2, handle, file_type)   
        os.system("cat %s %s > %s" % (err_free_subdataset_2, err_free_subdataset, true_data))     
        return raw_data, true_data

    def mutate_seq(self, shared_objects, idx):
        records_dict, file_type = shared_objects
        ori_seq = str(records_dict[idx].seq)
        mutation_seq = self.read2err_read(ori_seq)
        if file_type == "fastq":
            return SeqRecord(Seq(mutation_seq), id=records_dict[idx].id, description=records_dict[idx].description, letter_annotations=records_dict[idx].letter_annotations)
        elif file_type == "fasta":
            return SeqRecord(Seq(mutation_seq).seq, id=records_dict[idx].id, description=records_dict[idx].description)           
    '''
    def extract_seq(self, shared_objects, name):
        records_dict, file_type = shared_objects
        if file_type == "fastq":
            return SeqRecord(records_dict[name].seq, id=records_dict[name].id, description=records_dict[name].description, letter_annotations=records_dict[name].letter_annotations)
        elif file_type == "fasta":
            return SeqRecord(records_dict[name].seq, id=records_dict[name].id, description=records_dict[name].description)

    def extract_seqs(self, num_workers, name_lst, data_set, sub_dataset):
        records_dict, file_type = self.parse_data_dict(data_set)
        sub_records = []
        shared_objects = records_dict, file_type
        with WorkerPool(num_workers, shared_objects=shared_objects, start_method='fork') as pool:
            with tqdm(total=len(name_lst), desc=self.logger.info("Extract Seqs")) as pbar:   
                for tmp_rec in pool.imap(self.extract_seq, name_lst):
                    # print(tmp_rec)
                    sub_records.append(tmp_rec)
                    pbar.update() 
        # print(type(sub_records[0]), sub_records[0])
        with open(sub_dataset, "w") as handle:
            SeqIO.write(sub_records, handle, file_type)
        return 
    '''
    
# if __name__ == '__main__':
#     logger = custom_logger("simulation", debug_mode=True)
#     input_file = "/home/pping/Data/Repo/data/srrec_data/group3/SAS-Cov-2/srrec_result/new_no_ml/output/umi_simulated_raw_original_sars_r1_corrected_correct.fastq"
#     sim = Simulation(logger, 30, )

#     raw_data, true_data = sim.error_injection(input_file)
#     output_dir = "/home/pping/Data/Repo/data/srrec_data/group3/SAS-Cov-2/"
#     DP = DataProcessing( 
#         logger,
#         num_workers=30,
#         output_dir=output_dir)
#     umi_raw_dataset, umi_true_dataset = DP.write_mimic_umis(raw_data, true_data)